from . import conf, constants

if conf.ENABLE_MENU_REGISTRATION:
    from cms import constants as cms_constants
    from cms.apphook_pool import apphook_pool
    from cms.cms_menus import CMSMenu as OriginalCMSMenu, get_visible_nodes
    from cms.models import Page

    try:
        from cms.models import TreeNode
    except ImportError:
        TreeNode = None
    from cms.toolbar.utils import get_object_preview_url, get_toolbar_from_request
    from cms.utils.page import get_page_queryset
    from django.apps import apps
    from menus.base import Menu, NavigationNode
    from menus.menu_pool import menu_pool

    class CMSVersionedNavigationNode(NavigationNode):
        def is_selected(self, request):
            try:
                page_id = request.current_page.pk
            except AttributeError:
                return False
            return page_id == self.id

    def _get_attrs_for_node(renderer, page_content):
        page = page_content.page
        language = renderer.request_language
        attr = {
            "is_page": True,
            "soft_root": page_content.soft_root,
            "auth_required": page.login_required,
            "reverse_id": page.reverse_id,
        }
        limit_visibility_in_menu = page_content.limit_visibility_in_menu

        if limit_visibility_in_menu is cms_constants.VISIBILITY_ALL:
            attr["visible_for_authenticated"] = True
            attr["visible_for_anonymous"] = True
        else:
            attr["visible_for_authenticated"] = limit_visibility_in_menu == cms_constants.VISIBILITY_USERS
            attr["visible_for_anonymous"] = limit_visibility_in_menu == cms_constants.VISIBILITY_ANONYMOUS

        attr["is_home"] = page.is_home
        extenders = []

        if page.navigation_extenders:
            if page.navigation_extenders in renderer.menus:
                extenders.append(page.navigation_extenders)
            elif f"{page.navigation_extenders}:{page.pk}" in renderer.menus:
                extenders.append(f"{page.navigation_extenders}:{page.pk}")

        if page.application_urls:
            app = apphook_pool.get_apphook(page.application_urls)

            if app:
                extenders += app.get_menus(page, language)

        exts = []

        for ext in extenders:
            if hasattr(ext, "get_instances"):
                exts.append(f"{ext.__name__}:{page.pk}")
            elif hasattr(ext, "__name__"):
                exts.append(ext.__name__)
            else:
                exts.append(ext)

        if exts:
            attr["navigation_extenders"] = exts

        attr["redirect_url"] = page_content.redirect

        return attr

    class CMSMenu(Menu):
        """This is a legacy class used by django CMS 4.0 and django CMS 4.1.0 only. Its language
        fallback mechanism does not comply with django CMS' core's. Also, it is by far slower
        than django CMS core's. As of django CMS 4.1.1, this class is by default deactivated.

        See https://discord.com/channels/800813886689247262/1204047551570120755 for more information."""

        def get_nodes(self, request):
            site = self.renderer.site
            language = self.renderer.request_language
            pages_qs = get_page_queryset(site).select_related("node")
            visible_pages_for_user = get_visible_nodes(request, pages_qs, site)

            if not visible_pages_for_user:
                return []

            cms_extension = apps.get_app_config("djangocms_versioning").cms_extension
            toolbar = get_toolbar_from_request(request)
            edit_or_preview = toolbar.edit_mode_active or toolbar.preview_mode_active
            menu_nodes = []
            node_id_to_page = {}
            homepage_content = None

            # Depending on the toolbar mode, we need to get the correct version.
            # On edit or preview mode: return DRAFT,
            # if DRAFT does not exist then return PUBLISHED.
            # On public mode: return PUBLISHED.
            if edit_or_preview:
                states = [constants.DRAFT, constants.PUBLISHED]
            else:
                states = [constants.PUBLISHED]

            versionable_item = cms_extension.versionables_by_grouper[Page]
            versioned_page_contents = (
                versionable_item.content_model._base_manager.filter(
                    language=language, page__in=pages_qs, versions__state__in=states
                )
                .order_by("page__node__path" if TreeNode else "page__path", "versions__state")
                .select_related("page", "page__node" if TreeNode else "page")
                .prefetch_related("versions")
            )
            added_pages = []

            for page_content in versioned_page_contents:
                page = page_content.page

                if page not in visible_pages_for_user:
                    # The page is restricted for the user.
                    # Therefore, we avoid adding it to the menu.
                    continue

                version = page_content.versions.all()[0]

                if page.pk in added_pages and edit_or_preview and version.state == constants.PUBLISHED:
                    # Page content is already added. This is the case where you
                    # have both draft and published and in edit/preview mode.
                    # We give priority to draft which is already sorted by the query.
                    # Therefore we ignore the published version.
                    continue

                page_tree_node = page.node
                parent_id = node_id_to_page.get(page_tree_node.parent_id)

                if page_tree_node.parent_id and not parent_id:
                    # If the parent page is not available,
                    # we skip adding the menu node.
                    continue

                # Construct the url based on the toolbar mode.
                if edit_or_preview:
                    url = get_object_preview_url(page_content)
                else:
                    url = page_content.get_absolute_url()

                # Create the new navigation node.
                new_node = CMSVersionedNavigationNode(
                    id=page.pk,
                    attr=_get_attrs_for_node(self.renderer, page_content),
                    title=page_content.menu_title or page_content.title,
                    url=url,
                    visible=page_content.in_navigation,
                )

                if not homepage_content:
                    # Set the home page content.
                    homepage_content = page_content if page.is_home else None

                cut_homepage = homepage_content and not homepage_content.in_navigation

                if cut_homepage and parent_id == homepage_content.page.pk:
                    # When the homepage is hidden from navigation,
                    # we need to cut all its direct children from it.
                    new_node.parent_id = None
                else:
                    new_node.parent_id = parent_id

                node_id_to_page[page_tree_node.pk] = page.pk
                menu_nodes.append(new_node)
                added_pages.append(page.pk)
            return menu_nodes

    # Remove the core djangoCMS CMSMenu and register the new CMSVersionedMenu.
    menu_pool.menus.pop(OriginalCMSMenu.__name__)
    menu_pool.register_menu(CMSMenu)



import copy

from cms import api
from cms.models import PageContent, Placeholder
from cms.test_utils.testcases import CMSTestCase
from django.apps import apps
from django.test import TestCase

from djangocms_versioning.constants import ARCHIVED, PUBLISHED
from djangocms_versioning.datastructures import VersionableItem, default_copy
from djangocms_versioning.models import Version
from djangocms_versioning.test_utils.factories import PageContentFactory, PollVersionFactory
from djangocms_versioning.test_utils.people.models import PersonContent
from djangocms_versioning.test_utils.polls.models import Poll, PollContent

if not hasattr(CMSTestCase, "assertQuerySetEqual"):
    # Django < 4.2
    CMSTestCase.assertQuerySetEqual = CMSTestCase.assertQuerysetEqual


class VersionableItemTestCase(CMSTestCase):
    def setUp(self):
        self.initial_version = PollVersionFactory()

    def test_distinct_groupers(self):
        latest_poll1_version = PollVersionFactory(
            content__poll=self.initial_version.content.poll
        )
        poll2_version = PollVersionFactory()
        PollVersionFactory(content__poll=poll2_version.content.poll)
        latest_poll2_version = PollVersionFactory(
            content__poll=poll2_version.content.poll
        )

        versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )
        self.assertQuerySetEqual(
            versionable.distinct_groupers(),
            [latest_poll1_version.content.pk, latest_poll2_version.content.pk],
            transform=lambda x: x.pk,
            ordered=False,
        )

    def test_queryset_filter_for_distinct_groupers(self):
        poll1_archived_version = PollVersionFactory(
            content__poll=self.initial_version.content.poll, state=ARCHIVED
        )
        poll1_published_version = PollVersionFactory(
            content__poll=self.initial_version.content.poll, state=PUBLISHED
        )
        poll2_version = PollVersionFactory()
        PollVersionFactory(content__poll=poll2_version.content.poll, state=ARCHIVED)
        poll2_archived_version = PollVersionFactory(
            content__poll=poll2_version.content.poll, state=ARCHIVED
        )

        versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )

        qs_published_filter = {"versions__state__in": [PUBLISHED]}
        # Should be one published version
        self.assertQuerySetEqual(
            versionable.distinct_groupers(**qs_published_filter),
            [poll1_published_version.content.pk],
            transform=lambda x: x.pk,
            ordered=False,
        )

        qs_archive_filter = {"versions__state__in": [ARCHIVED]}
        # Should be two archived versions
        self.assertQuerySetEqual(
            versionable.distinct_groupers(**qs_archive_filter),
            [poll1_archived_version.content.pk, poll2_archived_version.content.pk],
            transform=lambda x: x.pk,
            ordered=False,
        )

    def test_for_grouper(self):
        poll1_version2 = PollVersionFactory(
            content__poll=self.initial_version.content.poll
        )
        poll2_version = PollVersionFactory()
        PollVersionFactory(content__poll=poll2_version.content.poll)
        PollVersionFactory(content__poll=poll2_version.content.poll)

        versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )

        self.assertQuerySetEqual(
            versionable.for_grouper(self.initial_version.content.poll),
            [self.initial_version.content.pk, poll1_version2.content.pk],
            transform=lambda x: x.pk,
            ordered=False,
        )

    def test_grouper_model(self):
        versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )

        self.assertEqual(versionable.grouper_model, Poll)

    def test_content_model_is_sideframe_editable_for_sideframe_disabled_model(self):
        """
        A content model with placeholders should not be opened in the sideframe
        """
        versionable = VersionableItem(
            content_model=PageContent,
            grouper_field_name="page",
            copy_function=default_copy,
        )

        self.assertEqual(versionable.content_model_is_sideframe_editable, False)

    def test_content_model_is_sideframe_editable_for_sideframe_enabled_model(self):
        """
        A content model without placeholders should be opened in the sideframe
        """
        versionable = VersionableItem(
            content_model=PollContent,
            grouper_field_name="poll",
            copy_function=default_copy,
        )

        self.assertEqual(versionable.content_model_is_sideframe_editable, True)


class VersionableItemProxyModelTestCase(CMSTestCase):
    @classmethod
    def setUpClass(cls):
        cls._all_models = copy.deepcopy(apps.all_models)

    @classmethod
    def tearDownClass(cls):
        apps.all_models = cls._all_models

    def tearDown(self):
        apps.all_models.pop("djangocms_versioning", None)

    def test_version_model_proxy(self):
        versionable = VersionableItem(
            content_model=PersonContent,
            grouper_field_name="person",
            copy_function=default_copy,
        )
        version_model_proxy = versionable.version_model_proxy

        self.assertIn(Version, version_model_proxy.mro())
        self.assertEqual(version_model_proxy.__name__, "PersonContentVersion")
        self.assertEqual(version_model_proxy._source_model, PersonContent)
        self.assertTrue(version_model_proxy._meta.proxy)

    def test_version_model_proxy_cached(self):
        """Test that version_model_proxy property is cached
        and return value is created once."""
        versionable = VersionableItem(
            content_model=PersonContent,
            grouper_field_name="person",
            copy_function=default_copy,
        )

        self.assertEqual(
            id(versionable.version_model_proxy), id(versionable.version_model_proxy)
        )

class DefaultCopyTestCase(TestCase):
    def setUp(self):
        self.original_content = PageContentFactory()

    def test_default_copy_creates_new_instance(self):
        new_content = default_copy(self.original_content)
        self.assertNotEqual(self.original_content.pk, new_content.pk)
        self.assertEqual(self.original_content.page, new_content.page)
        self.assertEqual(self.original_content.language, new_content.language)

    def test_default_copy_copies_placeholders(self):
        placeholder = Placeholder.objects.create(slot="content")
        self.original_content.placeholders.add(placeholder)
        new_content = default_copy(self.original_content)
        self.assertEqual(new_content.placeholders.count(), 1)
        self.assertNotEqual(new_content.placeholders.first().pk, placeholder.pk)
        self.assertEqual(new_content.placeholders.first().slot, placeholder.slot)

    def test_default_copy_copies_plugins_within_placeholder(self):
        # Create a placeholder and attach two different plugin types
        placeholder = Placeholder.objects.create(slot="content")
        plugin1 = api.add_plugin(
            placeholder=placeholder,
            plugin_type="TextPlugin",
            language=self.original_content.language,
            body="Sample text",
        )
        plugin2 = api.add_plugin(
            placeholder=placeholder,
            plugin_type="TextPlugin",
            language=self.original_content.language,
            body="Some other text",
        )
        self.original_content.placeholders.add(placeholder)

        new_content = default_copy(self.original_content)
        new_placeholder = new_content.placeholders.first()

        # Ensure that the new placeholder has two plugins
        self.assertEqual(new_placeholder.cmsplugin_set.count(), 2)

        # Collect original and copied plugin IDs for comparison
        original_plugin_ids = {plugin1.pk, plugin2.pk}
        new_plugins = list(new_placeholder.cmsplugin_set.all())
        for plugin in new_plugins:
            self.assertNotIn(plugin.pk, original_plugin_ids)

        # Verify that the copied plugins preserve type and key attributes
        downcasted = [plugin.get_plugin_instance()[0] for plugin in new_plugins]
        original = [plugin1, plugin2]
        for orig_plugin, new_plugin in zip(original, downcasted):
            self.assertEqual(orig_plugin.plugin_type, new_plugin.plugin_type)
            self.assertEqual(orig_plugin.body, new_plugin.body)

    def test_default_copy_copies_multiple_placeholders(self):
        placeholders = [Placeholder.objects.create(slot=f"slot_{i}") for i in range(3)]
        for placeholder in placeholders:
            self.original_content.placeholders.add(placeholder)
        new_content = default_copy(self.original_content)
        self.assertEqual(new_content.placeholders.count(), len(placeholders))
        for original in self.original_content.placeholders.all():
            copied = new_content.placeholders.get(slot=original.slot)
            self.assertNotEqual(copied.pk, original.pk)
            self.assertEqual(copied.slot, original.slot)

    def test_default_copy_calls_copy_relations_if_exists(self):
        class MockContent(PageContent):
            class Meta:
                app_label = "cms"
                proxy = True

            def __init__(self, *args, **kwargs):
                self.copy_relations_called = False
                super().__init__(*args, **kwargs)

            def copy_relations(self):
                self.copy_relations_called = True

        original_content = MockContent(language=self.original_content.language, page=self.original_content.page)
        new_content = default_copy(original_content)
        self.assertTrue(new_content.copy_relations_called)
